#include "FishUtils.H"
#include "dmartin/random.h"
#include "Structures.H"
#include <fstream.h>
#include "dmartin/tablet.h"

int g_i_id;

int gNumberFish = 50;
int gNumberNodes = 1;
int gDisplay = 1;
double gDtMin = 0.0001;
double gDtMax = 0.01;
double gDxMin = .005;
int gWindowSize = 200;
double gEndTime = 50;
double gTraceInterval = 1;
double gDomainSize = 2.0;
double gCutoff = 0.1;
int gCalcEnergy = 1;

char *gTraceFileName = "trace.out";

double gInvCutoff;
int gGridWidth;
double gSqrCutoff;

void usage (const char* pc_prog)
{
  if (g_i_id == 0)
    {
      fprintf (stderr, 
               "usage: %s "
               "-fish <n> "
               "-procs <n> "
               "-dtmin <x> "
               "-dtmax <x> "
               "-dxmin <x> "
               "-grav <x> "
               "-end <x> "
               "-display <yes|no> "
               "-size <n>"
               "-tint <x> "
               "-tname <s> "
               "-domain <x> "
               "-cutoff <x> "
               "-energy <yes|no> "
               "\n", pc_prog);
    }
  exit (1);
}

void badParam (const char* pc_prog)
{
  if (g_i_id == 0)
    {
      fprintf (stderr, "%s: bad parameter value\n", pc_prog);
    }
  exit (1);
}

int intArg (char* pc_prog, int* argc, char*** argv)
{
  int ret;
  (*argc)--; (*argv)++;
  if (*argc == 0) usage (pc_prog);
  ret = atoi (**argv);
  (*argc)--; (*argv)++;
  return ret;
}

double floatArg (char* pc_prog, int* argc, char*** argv)
{
  double ret;
  (*argc)--; (*argv)++;
  if (*argc == 0) usage (pc_prog);
  ret = atof (**argv);
  (*argc)--; (*argv)++;
  return ret;
}

char *stringArg(char* pc_prog, int *argc, char ***argv)
{
  char *ret;
  (*argc)--; (*argv)++;
  if (*argc == 0) usage (pc_prog);
  ret = **argv;
  (*argc)--; (*argv)++;
  return ret;
}

int boolArg (char* pc_prog, int* argc, char*** argv)
{
  bool ret;
  (*argc)--; (*argv)++;
  if (*argc == 0) usage (pc_prog);
  if (strcmp (**argv, "yes") == 0)
    {
      ret = 1;
    }
  else if (strcmp (**argv, "no") == 0)
    {
      ret = 0;
    }
  else
    {
      usage (pc_prog);
    }
  (*argc)--; (*argv)++;
  return ret;
}


int ProcessArguments(int argc, char **argv)
{
  char *pc_prog = *argv;
  
  argc--; argv++;
  while (argc > 0)
    {
      if (strcmp (*argv, "-fish") == 0) 
        {
          gNumberFish = intArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-procs") == 0) 
        {
          gNumberNodes = intArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-display") == 0) 
        {
          gDisplay = boolArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-dtmin") == 0) 
        {
          gDtMin = floatArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-dtmax") == 0) 
        {
          gDtMax = floatArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-dxmin") == 0) 
        {
          gDxMin = floatArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-end") == 0) 
        {
          gEndTime = floatArg (pc_prog, &argc, &argv);
        }
      else if (strcmp (*argv, "-size") == 0) 
        {
          gWindowSize = intArg (pc_prog, &argc, &argv);
        }
      else if (strcmp(*argv, "-tint") == 0)
        {
          gTraceInterval = floatArg(pc_prog, &argc, &argv);
        }
      else if (strcmp(*argv, "-tname") == 0)
        {
          gTraceFileName = stringArg(pc_prog, &argc, &argv);
        }
      else if (strcmp(*argv, "-domain") == 0)
        {
          gDomainSize = floatArg(pc_prog, &argc, &argv);
        }
      else if (strcmp(*argv, "-cutoff") == 0)
        {
          gCutoff = floatArg(pc_prog, &argc, &argv);
        }
      else if (strcmp(*argv, "-energy") == 0)
        {
          gCalcEnergy = boolArg(pc_prog, &argc, &argv);
        }

      else
        { 
          usage (pc_prog);
          return 0;
        }
    }

  // check argument values
  if (gNumberFish < 1 ||
      gNumberNodes < 1 ||
      gNumberNodes > gNumberFish ||
      gDtMin > gDtMax ||
      gDxMin < 0 ||
      gEndTime < 0 ||
      gWindowSize < 1) 
    {
      badParam (pc_prog);
      return 0;
    }

  return 1;
}


void ComputeDimensions()
{
  gInvCutoff = 1/gCutoff;
  gSqrCutoff = gCutoff * gCutoff;
  
  gGridWidth = 2 + (int) (gDomainSize * gInvCutoff);

  //cout << "Grid Width: " << gGridWidth << "\n";
  //cout << "Inverse Cutoff: " << gInvCutoff << "\n";
  
}

void InitializeFish(Fish **fa, int number)
{
  double rad = gDomainSize / 2.0;
  double vbase = rad / 100.0;
  
  Crand* p_rand = new Crand;
  p_rand->init (0U);
  for (int i = 0; i < number; i++)
    fa[i] = new Fish(p_rand->fp(0, 2*rad), p_rand->fp(0, 2*rad),
                     p_rand->fp (-vbase, vbase), p_rand->fp (-vbase, vbase),
                     i);
  
}

static Fish **print_buffer = NULL;

int fishcompare(Fish **a, Fish **b)
{
  Fish *fisha = (*a);
  Fish *fishb = (*b);

  double delta_px, delta_py, delta_vx, delta_vy;

  delta_px = fisha->px - fishb->px;
  delta_py = fisha->py - fishb->py;
  delta_vx = fisha->vx - fishb->vx;
  delta_vy = fisha->vy - fishb->vy;

  if (delta_px > 0)
    return 1;
  else if (delta_px < 0)
    return -1;
  else if (delta_py > 0)
    return 1;
  else if (delta_py < 0)
    return -1;
  else if (delta_vx > 0)
    return 1;
  else if (delta_vx < 0)
    return -1;
  else if (delta_vy > 0)
    return 1;
  else if (delta_vy < 0)
    return -1;
  else
    return 0; 
}

void PrintFishInfo(ofstream &tracefile, double time, Fish **fl, double energy)
{
  Fish *current;
  int i;
  
  if (print_buffer == NULL)
    print_buffer = new (Fish*)[gNumberFish];

  /* copy incoming fish over into the print_buffer...
     we'll use the print_buffer to resort the fish based on their
     instrinsic characteristics.. */

  for (i = 0; i < gNumberFish; i++)
    print_buffer[i] = fl[i];

  /* now sort the print_buffer using qsort */
  qsort( print_buffer, gNumberFish, sizeof(Fish *), fishcompare);
  
  tracefile << "Time: " << time << "  Energy: " << energy << "\n\n";
  
  for (i = 0; i < gNumberFish; i++)
    {
      current = print_buffer[i];
      tracefile << i << ": " << current->px << " " << current->py << " " <<
        current->vx << " " << current->vy << "\n";
    }
  tracefile << "\n";
}

void DrawFish(Ctablet *window, Fish *f)
{
  window->drawPoint( (f->px)/gDomainSize * 0.8 + 0.1,
                     (f->py)/gDomainSize * 0.8 + 0.1);
}


