#ifndef _FISH_UTILS_H
#define _FISH_UTILS_H

#include <stdlib.h>

///////////////////////////////////////////////////////////////////////
// MACROS
///////////////////////////////////////////////////////////////////////

#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))  
#define MIN(X,Y) ((X) > (Y) ? (Y) : (X))  
#define SQR(X)   ((X)*(X))
#define ABS(X)   ((X>=0) ? X : -(X))
#define DISTANCE(X,Y)  (sqrt((X)*(X) + (Y)*(Y)))

class Fish;
class ofstream;
class Ctablet;

///////////////////////////////////////////////////////////////////////
// GLOBALS
///////////////////////////////////////////////////////////////////////

extern int g_i_id;

extern int gNumberFish;
extern int gNumberNodes;
extern int gDisplay;
extern double gDtMin;
extern double gDtMax;
extern double gDxMin;
extern int gWindowSize;
extern double gEndTime;
extern double gTraceInterval;
extern double gDomainSize;
extern double gCutoff;
extern char *gTraceFileName;
extern int gCalcEnergy;

/* these are helper variables that pre-calced from inputted
   parameters */
extern double gInvCutoff;
extern int gGridWidth;
extern double gSqrCutoff;

///////////////////////////////////////////////////////////////////////
// ARGUMENT PROCESSING
///////////////////////////////////////////////////////////////////////

void usage (const char* pc_prog);
void badParam (const char* pc_prog);
int intArg (char* pc_prog, int* argc, char*** argv);
double floatArg (char* pc_prog, int* argc, char*** argv);
int boolArg (char* pc_prog, int* argc, char*** argv);
char *stringArg(char* pc_prog, int *argc, char ***argv);

int ProcessArguments(int argc, char **argv);

void ComputeDimensions();

void InitializeFish(Fish **fa, int number);

void PrintFishInfo(ofstream &tracefile, double time, Fish **fl, double energy);

void DrawFish(Ctablet *window, Fish *f);

#endif
