#include "List.H"

#ifndef _LIST_C_
#define _LIST_C_
#define DEBUG_LIST

template <class Ltype> Link<Ltype>::Link(Ltype *l)
  : data(l), next(NULL), previous(NULL), owner(NULL)
{
}

template <class Ltype> List<Ltype>::List(Link<Ltype> **lookup, int number)
  : myHead(NULL), myCurrent(NULL), myNumElements(0), myLookup(lookup),
    myMax(number)
{
}

template <class Ltype>List<Ltype>::~List()
{
}

template <class Ltype> void List<Ltype>::Add(Ltype *element)
{
#ifdef DEBUG_LIST
  int id = element->myId;
  if (id < 0)
    cout << "Error: element id less than 0\n";
  else if (id >= myMax)
    cout << "Error: element id greater than max\n";
#endif
  
  Link<Ltype> *temp = myLookup[element->myId];

#ifdef DEBUG_LIST
  if (temp->owner != NULL)
    {
      cout << "Error: trying to add an element that's already owned.\n";
      *((int *) 12) = 1;
    }
#endif

  if (myHead)
    myHead->previous = temp;

  temp->data = element;  
  temp->next = myHead;
  temp->previous = NULL;
  temp->owner = this;
  myHead = temp;
  
  myNumElements++;
  
}


template <class Ltype> void List<Ltype>::Remove(Ltype *element)
{
#ifdef DEBUG_LIST
  int id = element->myId;
  if (id < 0)
    cout << "Error: element id less than 0\n";
  else if (id >= myMax)
    cout << "Error: element id greater than max\n";
#endif
  
  Link<Ltype> *temp = myLookup[element->myId];

#ifdef DEBUG_LIST
  if (temp->owner != this)
    {
      cout << "Error: trying to remove an element that not in list.\n";
      *((int *) 12) = 1;
    }
#endif

  /* if temp has a previous link, then we must update that previous link's
     next to point past temp */
  if (temp->previous)
    temp->previous->next = temp->next;
  else /* if no previous, than it is the head, so update the head */
    myHead = temp->next;

  /* if temp has a next link, then update that next link's previous */
  if (temp->next)
    temp->next->previous = temp->previous;

  temp->owner = NULL;
  temp->next = NULL;
  temp->previous = NULL;

  myNumElements--;
  
}

template <class Ltype> void List<Ltype>::RemoveAll()
{
  Link<Ltype> *c = myHead;
  Link<Ltype> *t;
  
  while (c)
    {
      t = c->next;
      c->next = NULL;
      c->previous = NULL;
      c->owner = NULL;
      c = t;
    }

  myHead = NULL;
  myNumElements = 0;
}

template <class Ltype> void List<Ltype>::Reset()
{
  myCurrent = myHead;
}

template <class Ltype> Ltype *List<Ltype>::Next()
{
  Ltype *result = NULL;

  if (myCurrent)
    {
      result = myCurrent->data;
      myCurrent = myCurrent->next;
    }

  return result;
}

template <class Ltype> int List<Ltype>::NumElements()
{
  return myNumElements;
}

template <class Ltype> int List<Ltype>::Member(Ltype *element)
{
#ifdef DEBUG_LIST
  int id = element->myId;
  if (id < 0)
    cout << "Error: element id less than 0\n";
  else if (id >= myMax)
    cout << "Error: element id greater than max\n";
#endif
  
  Link<Ltype> *temp = myLookup[element->myId];

  return (temp->owner == this);
}

template <class Ltype> void List<Ltype>::TestPrint(ostream &os)
{
  /* First, check to make sure the doubly-linked structure is correct,
     and check to make sure the NumElements matches */
  Link<Ltype> *c = myHead;
  Link<Ltype> *previous = NULL;
  int count = 0;
  int correct = 1;

  while (c)
    {
      count++;
      if (c->previous != previous)
        correct = 0;
      previous = c;
      c = c->next;
    }

  if (!correct)
    cout << "Error:  The doubly-linked structure is wrong.\n";
  if (count != myNumElements)
    cout << "Error:  Incorrect count\n";

  os << "Elements: " << myNumElements << "\n";
  c = myHead;
  while (c)
    {
      os << "  ";
      c->data->TestPrint(os);
      c = c->next;
    }
  
}


template <class Ltype>ListGenerator<Ltype>::ListGenerator(Ltype **a, int number)
  : myMax(number)
{
  myLookup = new (Link<Ltype>*)[number];

  for (int i = 0; i < number; i++)
    myLookup[i] = new Link<Ltype>(a[i]);
}

template <class Ltype> List<Ltype> *ListGenerator<Ltype>::GenerateList()
{
  return new List<Ltype>(myLookup, myMax);
}

#endif
