#include "Structures.H"
#include "FishUtils.H"
#include <iostream.h>
#include <fstream.h>
#include "dmartin/tablet.h"
#include <math.h>
#include "mpi.h"
#include "dmartin/timer.h"
#include "List.C"

void Build_MPI_Fish(MPI_Datatype *datatype);
void CollectFishInfo(Node *mainNode, MPI_Datatype &MPI_Fish);

int main(int argc, char **argv)
{
  MPI_Datatype MPI_Fish;
  MPI_Request *send_requests;
  MPI_Request *receive_requests;
  MPI_Status *send_status, *receive_status;
  Node **neighbors;
  int num_neighbors;
  Fish **mainFishList;
  Node *mainNode;
  ofstream tracefile;
  int stage1Tag = 30;
  int stage2Tag = 76;
  double time;
  double timeDelta;
  double maxSqrVel, myMaxSqrVel;
  double nextTraceTime = gTraceInterval;
  double energy;
  double innertemp;
  Ctablet *window = NULL;
  Node *currentNeighbor;
  Ctimer total_time;
  Ctimer comm_time;
  Ctimer calc_time;
  Ctimer buffer_time;
  Ctimer wait_time;
  Ctimer wait_time1;
  Ctimer wait_time2;
  Ctimer wait_time3;
  Ctimer wait_time4;
  Ctimer reduce_time;
  
  int i;

  /* Initialize MPI, and all that */
  MPI_Init (&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &g_i_id);
  MPI_Comm_size(MPI_COMM_WORLD, &gNumberNodes);

  if (!ProcessArguments(argc, argv))
    return -1;

  /* Tell MPI about my FishData structure */
  Build_MPI_Fish(&MPI_Fish);

  /* Configure outselves for the given domain size and cutoff */
  ComputeDimensions();

  //if (g_i_id == 0)
  //  tracefile.open(gTraceFileName);
  
  /* create the main node .. must pass into it a list of all the fish,
     after they have been initialized with the global allocation feature */

  mainFishList = new (Fish*)[gNumberFish];
  InitializeFish(mainFishList, gNumberFish);
  mainNode = new Node(g_i_id, gNumberNodes, gNumberFish, mainFishList);

  /* Now, let's setup the neighbor information */

  num_neighbors = mainNode->mainNeighbors->NumElements();

  neighbors = new (Node *)[num_neighbors];
  send_requests = (MPI_Request *) malloc(sizeof(MPI_Request) * num_neighbors);
  receive_requests = (MPI_Request *)
    malloc(sizeof(MPI_Request) * num_neighbors);
  send_status = (MPI_Status *) malloc(sizeof(MPI_Status) * num_neighbors);
  receive_status = (MPI_Status *) malloc(sizeof(MPI_Status) * num_neighbors);
  
  mainNode->mainNeighbors->Reset();
  i = 0;
  while ((currentNeighbor = mainNode->mainNeighbors->Next()))
    neighbors[i++] = currentNeighbor;
  
  /* Do some display crude */
  if (gDisplay && (g_i_id == 0))
    {
      window = new  Ctablet();
      window->init("Fish!", gWindowSize, gWindowSize);
    }

  //if (g_i_id == 0)
  //  PrintFishInfo(tracefile, 0.0, mainNode->currentFish, 0.0);

  total_time.reset();
  comm_time.reset();
  calc_time.reset();
  buffer_time.reset();
  wait_time.reset();
  
  total_time.start();
  
  /* Now do the darn simulation */
  for (timeDelta = gDtMin, time = 0.0; time < gEndTime; time += timeDelta)
    {
      maxSqrVel = -1.0;
      energy = 0.0;
      
      /* Stage one:  push all the fish to new nodes */
      /* First, step for all the receives */

      for (i = 0; i < num_neighbors; i++)
        MPI_Irecv( neighbors[i]->incomingBuffer, gNumberFish + 1,
                   MPI_Fish, neighbors[i]->myId, stage1Tag,
                   MPI_COMM_WORLD, &(receive_requests[i]));


      calc_time.start();
      mainNode->SetupForCalculations();
      calc_time.stop();

      myMaxSqrVel = 0.0;
      
      for (i = 0; i < num_neighbors; i++)
        {
          Node **ordering;
          int count;

          calc_time.start();

          ordering = mainNode->myStage1Ordering;
      
          innertemp =
            mainNode->PerformCalcsForList(timeDelta, time, energy,
                                          ordering[i]->myStage1Dependencies);

          myMaxSqrVel = MAX(myMaxSqrVel, innertemp);    
          calc_time.stop();

          /* Now we are ready to send that neighbor's stage one buffer */
          comm_time.start();
          buffer_time.start();
          mainNode->FillStageOneFor(ordering[i]->myId);
          buffer_time.stop();

          count = ordering[i]->outgoingBuffer[0].fish_num;

          MPI_Isend( ordering[i]->outgoingBuffer, count,
                     MPI_Fish, ordering[i]->myId, stage1Tag,
                     MPI_COMM_WORLD, &(send_requests[i]));

          comm_time.stop();
        }

      calc_time.start();
      innertemp = mainNode->PerformCalcsForList(timeDelta, time, energy,
                                                mainNode->myStage2Leftovers);
      myMaxSqrVel = MAX(myMaxSqrVel, innertemp);
      calc_time.stop();

      /* Now wait until everything is done */

      comm_time.start();
      wait_time.start();
      wait_time1.start();
      MPI_Waitall(num_neighbors, send_requests, send_status);
      wait_time1.stop();
      wait_time2.start();
      MPI_Waitall(num_neighbors, receive_requests, receive_status);
      wait_time.stop();
      wait_time2.stop();
      comm_time.stop();
      
      /* Receive the data */
      buffer_time.start();
      mainNode->ReceiveAllStageOne();
      buffer_time.stop();

      /* Stage Two:  give everyone their ghost info */
      
      for (i = 0; i < num_neighbors; i++)
        MPI_Irecv( neighbors[i]->incomingBuffer, gNumberFish + 1,
                   MPI_Fish, neighbors[i]->myId, stage2Tag,
                   MPI_COMM_WORLD, &(receive_requests[i]));
            
      for (i = 0; i < num_neighbors; i++)
        {
          Node **ordering;
          int count;

          calc_time.start();

          ordering = mainNode->myStage2Ordering;
      
          innertemp =
            mainNode->PerformCalcsForList(timeDelta, time, energy,
                                          ordering[i]->myStage2Dependencies);

          myMaxSqrVel = MAX(myMaxSqrVel, innertemp);
          calc_time.stop();

          /* Now we are ready to send that neighbor's stage one buffer */
          comm_time.start();
          buffer_time.start();
          mainNode->FillStageTwoFor(ordering[i]->myId);
          buffer_time.stop();

          count = ordering[i]->outgoingBuffer[0].fish_num;

          MPI_Isend( ordering[i]->outgoingBuffer, count,
                     MPI_Fish, ordering[i]->myId, stage2Tag,
                     MPI_COMM_WORLD, &(send_requests[i]));

          comm_time.stop();
        }

      calc_time.start();
      innertemp = mainNode->PerformCalcsForList(timeDelta, time, energy,
                                                mainNode->myStage1Leftovers);
      myMaxSqrVel = MAX(myMaxSqrVel, innertemp);
      calc_time.stop();
      
      /* Now wait until everything is done */

      comm_time.start();
      wait_time3.start();
      wait_time.start();
      MPI_Waitall(num_neighbors, send_requests, send_status);
      wait_time3.stop();
      wait_time4.start();
      MPI_Waitall(num_neighbors, receive_requests, receive_status);
      wait_time.stop();
      wait_time4.stop();
      comm_time.stop();
      
      /* Receive the data */
      buffer_time.start();
      mainNode->ReceiveAllStageTwo();
      buffer_time.stop();


      mainNode->FinalizeCalculations();
      
      if (gDisplay)
        {
          comm_time.start();
          /* Collect all the fish information */
          CollectFishInfo(mainNode, MPI_Fish);
          comm_time.stop();
          
          /* if the trace interval time was been tripped, print it out */
          if ((time > nextTraceTime) && (g_i_id == 0))
            {
              //PrintFishInfo(tracefile, time, mainNode->currentFish, energy);
              nextTraceTime += gTraceInterval;
            }
          
          /* Draw all the fish */
          if (g_i_id == 0)
            {
              for (i = 0; i < gNumberFish; i++)
                DrawFish(window, mainNode->currentFish[i]);
              
              window->renderScene();    
            }
        }

      comm_time.start();
      reduce_time.start();
      /* now get the darn reduced maxSqrVel from all the other nodes */
      MPI_Allreduce (&myMaxSqrVel, &maxSqrVel, 1, MPI_DOUBLE, MPI_MAX,
                     MPI_COMM_WORLD);
      reduce_time.stop();
      comm_time.stop();
      
      /* make sure to update the timeDelta */
      timeDelta = gDxMin / sqrt(maxSqrVel);
      timeDelta = MIN(gDtMax, timeDelta);
      timeDelta = MAX(gDtMin, timeDelta);
    }

  /* Just for the record, we get the last state of the fish */

  total_time.stop();

  CollectFishInfo(mainNode, MPI_Fish);

  if (g_i_id == 0)
    {
      //PrintFishInfo(tracefile, time, mainNode->currentFish, energy);
      //tracefile.close();

      cout << "\n";
      cout << "Total time: " << total_time.value() << "\n";
      cout << " Comm time: " << comm_time.value() << "\n";
      cout << " Calc time: " << calc_time.value() << "\n";
      cout << " Buff time: " << buffer_time.value() << "\n";
      cout << " Wait time: " << wait_time.value() << "\n";
      cout << " Wait1 time: " << wait_time1.value() << "\n";
      cout << " Wait2 time: " << wait_time2.value() << "\n";
      cout << " Wait3 time: " << wait_time3.value() << "\n";
      cout << " Wait4 time: " << wait_time4.value() << "\n";
      cout << " Reduce time: " << reduce_time.value() << "\n";
      
      cout << "\n";
    }
  
  
  MPI_Finalize();

  return 0;
}

void Build_MPI_Fish(MPI_Datatype *datatype)
{
  FishData *sample;
  int block_lengths[6];
  MPI_Aint displacements[6];
  MPI_Aint addresses[7];
  MPI_Datatype typelist[6];

  typelist[0] = MPI_INT;
  typelist[1] = MPI_INT;
  typelist[2] = MPI_DOUBLE;
  typelist[3] = MPI_DOUBLE;
  typelist[4] = MPI_DOUBLE;
  typelist[5] = MPI_DOUBLE;

  block_lengths[0] = 1;
  block_lengths[1] = 1;
  block_lengths[2] = 1;
  block_lengths[3] = 1;
  block_lengths[4] = 1;
  block_lengths[5] = 1;

  sample = (FishData *) malloc(sizeof(FishData));

  MPI_Address(sample, &addresses[0]);
  MPI_Address(&(sample->fish_num), &addresses[1]);
  MPI_Address(&(sample->grid_num), &addresses[2]);
  MPI_Address(&(sample->px), &addresses[3]);
  MPI_Address(&(sample->py), &addresses[4]);
  MPI_Address(&(sample->vx), &addresses[5]);
  MPI_Address(&(sample->vy), &addresses[6]);
  
  displacements[0] = addresses[1] - addresses[0];
  displacements[1] = addresses[2] - addresses[0];
  displacements[2] = addresses[3] - addresses[0];
  displacements[3] = addresses[4] - addresses[0];
  displacements[4] = addresses[5] - addresses[0];
  displacements[5] = addresses[6] - addresses[0];

  MPI_Type_struct(6, block_lengths, displacements, typelist,
                  datatype);

  MPI_Type_commit(datatype);

}


void CollectFishInfo(Node *mainNode, MPI_Datatype &MPI_Fish)
{
  int i;
  int ourTag2 = 33;
  MPI_Status status;
  
  if (g_i_id == 0)
    {
      /* then we must receive the information from all the other
         nodes */
      for (i = 1; i < gNumberNodes; i++)
        {
          MPI_Recv(mainNode->fishInfoInBuffer[i], gNumberFish + 1,
                   MPI_Fish, i, ourTag2, MPI_COMM_WORLD, &status);
          //cout << "Got fish from " << i << "\n";
        }
      mainNode->ReceiveFishInfoBuffers();
    }
  else
    {
      /* we must send our porition of the fish to the main node */
      mainNode->FillFishInfoBuffer();
      MPI_Send(mainNode->fishInfoOutBuffer,
               (int) mainNode->fishInfoOutBuffer[0].fish_num,
               MPI_Fish, 0, ourTag2, MPI_COMM_WORLD);
      //cout << "Sent fish to " << 0 << "\n";
    }
}




