#include "Structures.H"
#include "FishUtils.H"
#include <iostream.h>
#include <fstream.h>
#include "dmartin/tablet.h"
#include <math.h>

int main(int argc, char **argv)
{
  Fish **mainFishList;

  ofstream tracefile;
  double time;
  double timeDelta;
  double maxSqrVel;
  double nextTraceTime = gTraceInterval;
  Ctablet *window;
  double energy;
  
  int i, outer, inner;
  Fish *current;
  
  if (!ProcessArguments(argc, argv))
    return -1;

  ComputeDimensions();
  
  
  tracefile.open(gTraceFileName);
  
  /* create the main node list... Each of these nodes believes it
     is on a single processor.. we can trick them so that they will
     communicate to each other eventhough they are on the same processor */

  mainFishList = new (Fish*)[gNumberFish];
  InitializeFish(mainFishList, gNumberFish);
  

  if (gDisplay)
    {
      window = new  Ctablet();
      window->init("Fish!", gWindowSize, gWindowSize);
    }

  PrintFishInfo(tracefile, 0.0, mainFishList,0.0);
  
  for (timeDelta = gDtMin, time = 0.0; time < gEndTime; time += timeDelta)
    {
      energy = 0.0;
      
      maxSqrVel = -1.0;

      for (outer = 0; outer < gNumberFish; outer++)
        {
          current = mainFishList[outer];
          current->fx = 0.0;
          current->fy = 0.0;
          current->potential = 0.0;
          
          for (inner = 0; inner < gNumberFish; inner++)
            if (outer != inner)
              {
                current->CalcForce(mainFishList[inner]);
                if (gCalcEnergy)
                  current->CalcPotential(mainFishList[inner]);
              }
        }


      if (gCalcEnergy)
        for (outer = 0; outer < gNumberFish; outer++)
          {
            energy += (mainFishList[outer]->KineticEnergy() +
                       0.5 * mainFishList[outer]->potential);
          }
      
      for (outer = 0; outer < gNumberFish; outer++)
        {
          double temp = mainFishList[outer]->UpdateFish(timeDelta);
          maxSqrVel = MAX(temp, maxSqrVel);
        }

      if (gDisplay)
        {
          /* if the trace interval time was been tripped, print it out */
          if (time > nextTraceTime)
            {
              PrintFishInfo(tracefile, time, mainFishList, energy);
              nextTraceTime += gTraceInterval;
            }
          
          /* Draw all the fish */
          if (gDisplay)
            {
              for (i = 0; i < gNumberFish; i++)
                DrawFish(window, mainFishList[i]);
              
              window->renderScene();    
            }
        }
      
      /* make sure to update the timeDelta */
      timeDelta = gDxMin / sqrt(maxSqrVel);
      timeDelta = MIN(gDtMax, timeDelta);
      timeDelta = MAX(gDtMin, timeDelta);
    }

  PrintFishInfo(tracefile, time, mainFishList, energy);
  tracefile.close();
}
